/*
 * Copyright (C) 2006 Eskil Bylund
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

using System;
using System.Collections;
using System.Diagnostics;
using System.Text;

using Gtk;
using Glade;
using Mono.Unix;

using DCSharp.Backend.Connections;
using DCSharp.Backend.Objects;
using DCSharp.Backend.Managers;

namespace DCSharp.GUI
{
	public class ChatPage : Page
	{
		[Widget]
		private ChatView chatView;
		
		[Widget]
		private Entry messageEntry;
		
		#region Constructors
		
		public ChatPage(LocalIdentity localIdentity, Identity remoteIdentity) :
			base("ChatPage.glade")
		{
			if (localIdentity == null)
			{
				throw new ArgumentNullException("localIdentity");
			}
			if (remoteIdentity == null)
			{
				throw new ArgumentNullException("remoteIdentity");
			}
			
			this.localIdentity = localIdentity;
			this.remoteIdentity = remoteIdentity;
			
			chatView.Show();
			
			messageEntry.GrabFocus();
			messageEntry.Activated += OnMessageEntryActivated;
			
			// Events
			remoteIdentity.User.OnlineChanged += OnOnlineChanged;
		}
		
		#endregion
		
		#region Properties
		
		public override string Title
		{
			get { return remoteIdentity.Nick; }
		}
		
		public override Gdk.Pixbuf Icon
		{
			get { return Util.GetUserIcon(remoteIdentity); }
		}
		
		public override string Tooltip
		{
			get { return remoteIdentity.Nick; }
		}
		
		public Identity LocalIdentity
		{
			get { return localIdentity; }
		}
		private Identity localIdentity;
		
		public Identity RemoteIdentity
		{
			get { return remoteIdentity; }
		}
		private Identity remoteIdentity;
		
		#endregion
		
		#region Methods
		
		public override void Dispose()
		{
			remoteIdentity.User.OnlineChanged -= OnOnlineChanged;
			base.Dispose();
		}
		
		public void DisplayMessage(User from, string message)
		{
			if (from == null)
			{
				throw new ArgumentNullException("from");
			}
			if (message == null)
			{
				throw new ArgumentNullException("message");
			}
			if (from != localIdentity.User && from != remoteIdentity.User)
			{
				throw new ArgumentException("from", "Unknown user.");
			}
			
			Identity identity = from == localIdentity.User ? localIdentity :
				remoteIdentity;
			
			chatView.InsertMessage(identity.Nick, message, identity == localIdentity);
			OnPageChanged();
		}
		
		private void DisplayStatus()
		{
			string status = remoteIdentity.User.IsOnline ?
				Catalog.GetString("{0} has come online") :
				Catalog.GetString("{0} went offline");
			
			status = String.Format(status, remoteIdentity.Nick);
			
			chatView.InsertStatus(status);
			Status = status;
		}
		
		private void OnMessageEntryActivated(object o, EventArgs args)
		{
			string message = messageEntry.Text.Trim();
			if (message.Length > 0)
			{
				messageEntry.Text = String.Empty;
				
				HubConnection hub = Runtime.HubManager.GetHubWithUser(remoteIdentity.User);
				if (hub != null)
				{
					hub.SendPrivateMessage(remoteIdentity, message);
				}
			}
		}
		
		private void OnOnlineChanged(object obj, EventArgs args)
		{
			Application.Invoke(delegate
			{
				DisplayStatus();
				OnPageChanged();
			});
		}
		
		#endregion
	}
}
